import numpy as np  # type: ignore
import torch  # type: ignore
from torch.autograd import Variable  # type: ignore


class Net(torch.nn.Module):
    def __init__(self, n_in, n_hidden, n_out):
        super(Net, self).__init__()
        self.w1 = torch.nn.Linear(n_in, n_hidden)
        self.w2 = torch.nn.Linear(n_hidden, n_out)

    def forward(self, x):
        x = torch.tanh(self.w1(x))
        x = self.w2(x)
        return x

    def my_train(self, xtrain, ytrain, num_epochs):
        """
        Train the network

        Parameters
        ----------

        xtrain : np.ndarray
            Inputs

        ytrain : np.ndarray
            Corresponding desired outputs
        """

        # Convert from numpy arrays to torch tensors
        xtrain = Variable(torch.FloatTensor(xtrain))
        ytrain = Variable(torch.FloatTensor(ytrain))

        criterion = torch.nn.MSELoss(reduction="sum")
        optimizer = torch.optim.SGD(self.parameters(), lr=1e-5)

        for t in range(num_epochs):

            optimizer.zero_grad()

            y_pred = self(xtrain)
            loss = criterion(y_pred, ytrain)
            loss.backward()
            optimizer.step()

        print(t, loss.item() / xtrain.shape[0])

    def call_numpy(self, x: np.ndarray):
        """
        Call the network with numpy input and output
        """

        x_tensor = Variable(torch.FloatTensor(x))

        out = self(x_tensor)

        return out.detach().numpy()



from __future__ import print_function

from typing import Callable, List, Tuple, Union

import numpy as np  # type: ignore

from .emulator import BaseEmulator
from .util import check_good


def emulate(emulator_class) -> Callable:
    """
    Emulate a function with the given emulator class

    This function is intended to be used as a decorator on non-trivial
    functions.
    The function should take one numpy array as an argument and return one
    numpy array.

    Examples
    --------

    TODO: More thorough example

    >>> from layg import emulate, CholeskyNnEmulator
    >>> @emulate(CholeskyNnEmulator)
    >>> def expensive_calculation(x: np.ndarray) -> np.ndarray:
    >>>     # Function definition
    >>>     return np.ones_like(x)

    """

    assert issubclass(emulator_class, BaseEmulator)

    def inner(true_func: Callable) -> Learner:

        return Learner(true_func, emulator_class)

    return inner


class Learner(object):
    """
    A class that contains logic for learning as you go

    This class does not contain any emulation but should be constructed with an
    emulator containing emulation logic.
    The emulator must be a subclass of BaseEmulator, implementing two methods,
    `set_emul_func` and `set_emul_error_func`, that set the respective
    functions.

    Attributes
    ----------

    emulator_class : BaseEmulator
        The type of emulator used.

    emulator : BaseEmulator
        An instance of the class `emulator_class`.
        This is where the heavy lifting goes on.

    true_func : Callable
        The function which is emulated

    frac_err_local : float
        Maximum fractional error in emulated function.  Calls to emulation function
        that exceed this error level are evaluated exactly instead.
        Default: 1.0

    abs_err_local : float
        Maximum absolute error allowed in emulated function.  Calls to emulation function
        that exceed frac_err_local but are lower than abs_err_local are emulated, rather
        than exactly evaluated.
        FIXME: this doesn't happen
        Default: 0.05

    output_err : bool
        Whether to output an error estimate.
        Set to False if you do not want the error to be an output of the emulated function.
        Set to True if you do.
        Default: False

    trained : bool
        Whether the emulator has been trained

    used_train_x : List[np.ndarray]
    used_train_y : List[np.ndarray]
        Values from the true function that were used last time the emulator was
        trained

    batch_train_x : List[np.ndarray]
    batch_train_y : List[np.ndarray]
        Values from the true function that have not yet been used to train the
        emulator

    init_train_thresh : int
        Number of points to accumulate before training the emulator

    frac_cv : float
        Fraction of training set to use for error modelling
        The default value of 0.5 means that the prediction and the error are
        estimated off the same amount of data.
    """

    def __init__(self, true_func: Callable[[np.ndarray], np.ndarray], emulator_class):
        """
        Constructor for Learner class

        Parameters
        ----------
        true_func : Callable
            Function to be emulated

        emulator_class : BaseEmulator
            The emulator class to be used
        """

        self.emulator_class = emulator_class
        assert issubclass(self.emulator_class, BaseEmulator)

        self.emulator = self.emulator_class()

        self.true_func: Callable[[np.ndarray], np.ndarray] = true_func

        self.frac_err_local: float = 1.0
        self.abs_err_local: float = 0.05

        self.output_err: bool = False

        self.used_train_x: List[np.ndarray] = []
        self.used_train_y: List[np.ndarray] = []

        self.batch_train_x: List[np.ndarray] = []
        self.batch_train_y: List[np.ndarray] = []

        self._size_last_trained = np.nan

        self.init_train_thresh: int = 1000

        # Fraction of training set to use for error modelling
        self.frac_cv = 0.5

        # Whether the emulator is trained
        self.trained: bool = False
        self._num_times_trained: int = 0

        # Number of exact or emulated evaluations
        self._nexact: int = 0
        self._nemul: int = 0
        self._nemul_failed: int = 0

    def eval_true_func(self, x: np.ndarray) -> np.ndarray:
        """
        Wrapper for evaluating true function

        You want this so that you can do some pre-processing, training, or
        saving each time the emulator gets called.

        Train the emulator if the specified thresholds are met.

        Parameters
        ----------

        x : np.ndarray
            Independent variable.
            Assumed to be a set of vectors in :math:`R^n`.
        """

        myY: np.ndarray = self.true_func(x)

        # Add x, val to a batch list that we will hold around
        # TODO: lists should have no duplicates
        self.batch_train_x.append(x)
        self.batch_train_y.append(myY)

        # Add points to the emulator
        # No effect unless the emulator implements updating on the fly
        self.emulator.add_data(np.array([x]), np.array([myY]))

        retraining_threshold = (1 + self.frac_cv) * self._size_last_trained
        current_points = len(self.used_train_x) + len(self.batch_train_x)

        # Check if list size has increased above some threshold
        # If so, train for first time
        if not self.trained and len(self.batch_train_x) >= self.init_train_thresh:

            self.train(np.array(self.batch_train_x), np.array(self.batch_train_y))

            # Empty the batch
            self.used_train_x.extend(self.batch_train_x)
            self.used_train_y.extend(self.batch_train_y)
            self.batch_train_x = []
            self.batch_train_y = []

        # Check if list size has increased above retraining threshold
        # If so, train again
        elif self.trained and current_points >= retraining_threshold:

            new_train_x = np.array(self.used_train_x + self.batch_train_x)
            new_train_y = np.array(self.used_train_y + self.batch_train_y)

            self.train(new_train_x, new_train_y)

            # Empty the batch
            self.used_train_x.extend(self.batch_train_x)
            self.used_train_y.extend(self.batch_train_y)
            self.batch_train_x = []
            self.batch_train_y = []

        return myY

    def train(self, x_train: np.ndarray, y_train: np.ndarray):
        """Train a ML algorithm to replace true_func: X --> Y.  Estimate error model via cross-validation.

        Parameters
        ----------
        x_train : ndarray
            Independent variable of training set.  Assumed to be a set of
            vectors in :math:`R^n`

        y_train : ndarray
            Dependent variable of training set.  Assumed to be a set of vectors
            in :math:`R^m`, although it has limited functionality if m!=1.
        """

        print("RETRAINING!------------------------")

        # TODO: remove this
        if self.output_err is not False:
            # raise Exception('Do not currently have capability to output the error to the chain.')
            pass

        self._size_last_trained = x_train.shape[0]

        # Separate into training and cross-validation sets with 50-50 split so that
        # the prediction and the error are estimated off the same amount of data
        x_train, y_train, x_cv, y_cv = self.split_CV(x_train, y_train, self.frac_cv)

        # Set the emulator function by calling to the subclass's particular method
        self.emulator.set_emul_func(x_train, y_train)

        # Set the emulator function by calling to the subclass's particular method
        CV_y_err = y_cv - np.array([self.emulator.emul_func(x) for x in x_cv])[0]
        assert y_cv.shape == CV_y_err.shape  # Bizarre bugs if this isn't true
        self.emulator.set_emul_error_func(x_cv, CV_y_err)

        # Provide the cross validation values to the emulator
        self.emulator.add_data(x_cv, y_cv)

        self.trained = True
        self._num_times_trained += 1

    def split_CV(self, xdata: np.ndarray, ydata: np.ndarray, frac_cv: float):
        """Splits a dataset into a cross-validation and training set.  Shuffles the data.

        Parameters
        ----------
        xdata : ndarray
            Independent variable of dataset.  Assumed to be a set of vectors in R^n

        ydata : ndarray
            Dependent variable of dataset.  Assumed to be a set of vectors in R^m.

        frac_cv : scalar
            Fraction of dataset to be put into the cross-validation set.

        Results
        -------
        xtrain : ndarray
            Independent variable of training set.  Assumed to be a set of vectors in R^n

        ytrain : ndarray
            Dependent variable of training set.  Assumed to be a set of vectors in R^m.

        x_cv : ndarray
            Independent variable of cross-validation set.  Assumed to be a set of vectors in R^n

        y_cv : ndarray
            Dependent variable of cross-validation set.  Assumed to be a set of vectors in R^m.
        """

        # Separate into training and cross-validation sets
        num_cv = int(frac_cv * xdata.shape[0])

        rand_subset = np.arange(xdata.shape[0])
        np.random.shuffle(rand_subset)

        xdata = np.array([xdata[rand_index] for rand_index in rand_subset])
        ydata = np.array([ydata[rand_index] for rand_index in rand_subset])

        x_cv = xdata[-num_cv:]
        y_cv = ydata[-num_cv:]

        xtrain = xdata[0:-num_cv]
        ytrain = ydata[0:-num_cv]

        return xtrain, ytrain, x_cv, y_cv

    def emulation_is_valid(self, val: np.ndarray, err: float) -> bool:
        """
        Check if an emulated value is valid and likely accurate

        Check whether the val and error are valid values and whether the
        estimated error is acceptable.

        Parameters
        ----------

        val : np.ndarray
            Proposed value

        err : float
            Error estimate

        Returns
        -------

        bool
        """

        # Value and error should not be inf, nan, etc
        goodval: bool = check_good(val)
        gooderr: bool = check_good(err)

        # These error checks are ok even if gooderr is False; that case is caught later
        small_abs_err: bool = np.all(np.abs(err) < self.abs_err_local)
        small_frac_err: bool = np.all(np.abs(err / val) < self.frac_err_local)

        return goodval and gooderr and small_abs_err and small_frac_err

    def __call__(self, x: np.ndarray) -> Union[np.ndarray, Tuple[np.ndarray, float]]:
        """
        The method that is executed when the wrapped function is called

        Try to use an emulated version of `true_func`.  If this is not possible
        because the error estimate is too high or the emulator is not trained,
        fall back to the true function.

        Parameters
        ----------
        x : ndarray
            Independent variable.
            Assumed to be a set of vectors in :math:`R^n`.

            This is either passed directly to the true function or to an emulator.

        Results
        -------
        val : float
            The exact or emulated value of the true function.

        (val, err) : (np.adarray, float)
            The exact or emulated value of the true function and an error estimate.
            This version is returned is `return_err` is set to true on the class.
        """

        val: np.ndarray
        err: float

        if self.trained:
            # Note: in the paper, it is assumed that error checking can be
            # performed before the evaluation of the emulator.  We can't do
            # this as the error check uses value for the fractional error.
            val, err = self.emulator.emul_func(x), self.emulator.emul_error(x)

            if self.emulation_is_valid(val, err):
                # print("Emulated -------", val, err#, self.true_func(x))
                self._nemul += 1
            else:
                #    print("Exact evaluation -----------",goodval,gooderr)
                val = self.eval_true_func(x)
                err = 0.0
                self._nexact += 1
                self._nemul_failed += 1

        else:
            val, err = self.eval_true_func(x), 0.0
            self._nexact += 1

        if self.output_err:
            return val, float(err)
        else:
            return val



from __future__ import print_function

from typing import Callable

import numpy as np  # type: ignore
import scipy.interpolate as interp  # type: ignore

from .emulator import BaseEmulator


class InterpolationEmulator(BaseEmulator):
    """
    An emulator based on interpolation

    This emulator uses `scipy.interpolate.interp1d` to interpolate functions
    with scalar image and preimage.

    It returns infinity for values outside the trained domain.
    TODO: determine whether an exception should be raised instead. Then the
    emulator should evaluate the true function.
    """

    def interpolator(self, xdata: np.ndarray, ydata: np.ndarray):

        if ydata.shape[1] > 1:
            raise TypeError(
                "Cannot interpolate when range has higher dimensions than 1."
            )

        if xdata.shape[1] > 1:
            raise TypeError(
                "The interpolator is not yet set up for higher dimensions than ",
                xdata.ndim - 1,
            )

        if xdata.shape[0] != ydata.shape[0]:
            raise TypeError("The x and y data do not have the same number of elements.")

        # Reshape data from Nx1 matrix to N-vector
        xdata = xdata.T[0]
        ydata = ydata.T[0]

        interp_funct = interp.interp1d(xdata, ydata)
        xmin = np.min(xdata)
        xmax = np.max(xdata)

        def predict(x):
            # Reshape data from 1x1 matrix to scalar
            x = x.T[0]

            if x < xmin or x > xmax:
                pred = np.inf
            else:
                pred = interp_funct(x)
            return pred

        return predict

    def set_emul_func(self, x_train: np.ndarray, y_train: np.ndarray) -> None:
        self.emul_func: Callable[[np.ndarray], np.ndarray] = self.interpolator(
            x_train, y_train
        )

    def set_emul_error_func(self, x_cv: np.ndarray, y_cv_err: np.ndarray) -> None:
        """
        TODO: This error estimate is pretty bad; it fits the error as a function of x.
        Should be as a function of distance to nearest point.
        """
        self.emul_error: Callable[[np.ndarray], np.ndarray] = self.interpolator(
            x_cv, y_cv_err
        )



from .emulator import (
    BaseEmulator,
    CholeskyNnEmulator,
    InterpolationEmulator,
    TorchEmulator,
)
from .learner import Learner, emulate

__all__ = [
    "Learner",
    "emulate",
    "BaseEmulator",
    "CholeskyNnEmulator",
    "InterpolationEmulator",
    "TorchEmulator",
]



"""
An emulator that uses a simple neural network
"""
from typing import Callable

import numpy as np  # type: ignore
from scipy.optimize import curve_fit  # type: ignore

from .emulator import BaseEmulator
from .nn import Net


class TorchEmulator(BaseEmulator):
    """
    Class that uses pytorch to do emulation

    The Universal Approximation Theorem says that any Lebesgue integrable
    function can be approximated by a feed-forward network with sufficient
    layers of sufficient width.  It doesn't guarantee that we can train the
    network though.
    """

    def set_emul_func(self, x_train: np.ndarray, y_train: np.ndarray) -> None:

        self.x_train = x_train
        self.y_train = y_train

        n_input = x_train.shape[1]
        n_output = y_train.shape[1]
        n_hidden = 20

        # Define network with appropriate size
        self.net = Net(n_input, n_hidden, n_output)

        # Set network mode to training
        self.net.train()

        # Train network on given data
        self.net.my_train(x_train, y_train, 5000)

        # Set network mode to evaluation
        self.net.eval()

        self.emul_func = self.net.call_numpy

    def set_emul_error_func(self, x_cv: np.ndarray, y_cv_err: np.ndarray) -> None:
        """
        Fit a quadratic to the residuals and mean distance to nearby points
        """

        num_neighbors = 2 * x_cv.shape[1]

        dist_list = np.empty(x_cv.shape[0])
        # y_pred = np.empty(y_cv.shape)
        for i, x in enumerate(x_cv):
            distance_array = ((self.x_train - x) ** -2).sum(axis=1) ** -0.5
            dist_list[i] = np.mean(np.sort(distance_array)[num_neighbors:])
            # y_pred[i] = self.net.call_numpy(x)

        # TODO: choose another function to scalarify errors
        errors = (y_cv_err ** 2).sum(axis=1) ** 0.5

        def error_model(dist, a, b, c):
            """
            The error model function to be fit
            """

            return a + b * dist + c * dist ** 2

        popt, _ = curve_fit(
            error_model, dist_list, errors, bounds=(np.zeros(3), 1 / np.zeros(3))
        )

        def error(x):

            distance_array = ((self.x_train - x) ** -2).sum(axis=1) ** -0.5
            distance = np.mean(np.sort(distance_array)[num_neighbors:])

            return error_model(distance, *popt)

        self.emul_error: Callable[[np.ndarray], np.ndarray] = error



from .cholesky_nn_emulator import CholeskyNnEmulator
from .emulator import BaseEmulator
from .interpolation_emulator import InterpolationEmulator
from .torch_emulator import TorchEmulator

__all__ = [
    "BaseEmulator",
    "CholeskyNnEmulator",
    "InterpolationEmulator",
    "TorchEmulator",
]



"""
Module defining the class `Emulator`, from which emulators inherit
"""
from abc import ABC, abstractmethod
from typing import Callable

import numpy as np  # type: ignore


def raise_not_implemented_error():
    raise NotImplementedError()


class BaseEmulator(ABC):
    """
    Base class from which emulators should inherit

    This class is abstract.
    The child class must implement the marked methods.
    """

    def __init__(self):

        # The emulating emulations should not exist until the model is trained
        self.emul_func: Callable[np.ndarray, float] = raise_not_implemented_error
        self.emul_error: Callable[np.ndarray, float] = raise_not_implemented_error

    @abstractmethod
    def set_emul_func(self, x_train: np.ndarray, y_train: np.ndarray) -> None:
        pass

    @abstractmethod
    def set_emul_error_func(self, x_cv: np.ndarray, y_cv_err: np.ndarray) -> None:
        pass

    def add_data(self, x_train: np.ndarray, y_train: np.ndarray) -> None:
        """
        Add data to the training set on the fly

        By default this method does nothing.

        Overriding this method is not mandatory (some emulators require a long
        time to train, so adding a couple new points may not make sense).

        Parameters
        ----------

        x_train : np.ndarray
            Array of x-values
        y_train : np.ndarray
            Array of y-values
        """

        pass



"""
Functions that are used in various places
"""

import math

import numpy as np  # type: ignore


def check_good(x):
    """Tests if scalar is infinity, NaN, or None.

    Parameters
    ----------
    x : scalar or numpy.ndarray
        Input to test.

    Results
    -------
    good : logical
        False if x is or contains inf, NaN, or None; True otherwise.
    """

    if type(x) == np.ndarray:
        if np.all(np.isfinite(x)):
            return True
        else:
            return False

    else:
        if x == np.inf or x == -np.inf or x is None or math.isnan(x):
            return False
        else:
            return True



# -*- coding: utf-8 -*-
"""Setup file for layg package

Developer install
-----------------
Run following command in prompt/terminal:
    pip install -e .
"""
import setuptools  # type: ignore

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="layg",  # name of package on import
    version="0.0.1",
    description="Learn-as-you-go emulator with error estimation",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/auckland-cosmo/LearnAsYouGoEmulator",
    author="Nathan Musoke, Layne Price",
    author_email="n.musoke@aucland.ac.nz, layne.c.price@gmail.com",
    license="Apache Licence (2.0)",
    packages=setuptools.find_packages(),
    install_requires=[
        "matplotlib",
        "numpy",
        "numpydoc",
        "torch",
        "scipy",
        "sphinx",
        "sphinx_rtd_theme",
        "emcee>=2,<3",
        "gif==1.0.3",
    ],  # dependencies
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    python_requires=">=3.5",
)



from __future__ import print_function

import warnings
from itertools import product
from typing import Callable

import numpy as np  # type: ignore
import scipy.optimize as opt  # type: ignore
from scipy import special  # type: ignore
from scipy.spatial import cKDTree as KDTree  # type: ignore

from ..util import check_good
from .emulator import BaseEmulator


class CholeskyNnEmulator(BaseEmulator):
    """
    An emulator based on Cholesky decomposition and nearest neighbours

    This emulator described in detail in arXiv:1506.01079.
    """

    def set_emul_func(self, x_train: np.ndarray, y_train: np.ndarray) -> None:
        self.emulator = cholesky_NN(x_train, y_train)
        self.emul_func = self.emulator.emulate

    def set_emul_error_func(self, x_cv: np.ndarray, y_cv_err: np.ndarray) -> None:
        self.emul_error: Callable[
            [np.ndarray], np.ndarray
        ] = self.emulator.train_dist_error_model(x_cv, y_cv_err)

    # TODO: override add_data


class cholesky_NN(object):
    """
    Class containing the main logic for the Cholesky and nearest neighbour based emulator
    """

    def __init__(self, xdata: np.ndarray, ydata: np.ndarray):
        """
        Initialise emulator with data

        Parameters
        ----------

        xdata : np.ndarray
            Data for the independent variable

        ydata : np.ndarray
            corresponding data for the dependent variable
        """

        # Do some tests here

        # Find data covariance
        cov = np.cov(xdata.T)

        if cov.shape == () and xdata.shape[1] == 1:
            # Handle input for a function on scalars
            # np.cov returns a scalar if xdata.T is a column vector, but
            # np.linalg.cholesky requires a 2-d array.
            # TODO: check more formally that this change makes sense
            cov = np.array([[cov]])

        # Cholesky decompose to make new basis
        L_mat = np.linalg.cholesky(cov)
        self.L_mat = np.linalg.inv(L_mat)

        # Transform xdata into new basis
        self.xtrain = xdata
        self.transf_x = np.array([np.dot(self.L_mat, x) for x in xdata])

        # DEBUG
        # import matplotlib.pyplot as plt  # type: ignore
        # plt.plot(xdata[:,0],xdata[:,1],'.',color='r')
        # plt.plot(self.transf_x[:,0],self.transf_x[:,1],'.')
        # plt.show()
        # import sys; sys.exit()

        # Store training
        self.ytrain = ydata

        # Build KDTree for quick lookup
        self.transf_xtree = KDTree(self.transf_x)

    def emulate(self, x, delta=2):

        if delta < 0 or not isinstance(delta, int):
            raise ValueError(
                "Got degree of interpolation {}, must be a non-negative integer".format(
                    delta
                )
            )
        if x.ndim != self.xtrain[0].ndim:
            raise Exception(
                "Requested x and training set do not have the same number of dimension."
            )

        # Change basis
        x0 = np.dot(self.L_mat, x)

        # Need enough neighbours to constrain fit with degree delta polynomials
        k = 2 * n_coef(x.shape[0], delta)

        # Get nearest neighbors
        dist, loc = self.transf_xtree.query(x0, k=k)
        nearest_x = self.transf_x[loc]
        nearest_y = self.ytrain[loc]

        # Protect div by zero
        dist = np.array([np.max([1e-15, d]) for d in dist])
        weight = 1.0 / dist

        try:
            y_predict = polynomial_fit(x0, nearest_x, nearest_y, weight, delta)
        except np.linalg.LinAlgError as lin_alg_error:
            # TODO: this is a hack; should propagate exceptions carefully
            warnings.warn("Exception inside emulator: {}".format(lin_alg_error))

            # return nans and let caller deal with it
            output = np.empty(self.ytrain[0].shape[0])
            output.fill(np.nan)
            return output

        if not check_good(y_predict):
            raise Exception("y prediction went wrong")

        return y_predict

    def train_dist_error_model(self, xtrain, ytrain, k=5):
        """Rather than learning a non-parametric error model, we can define a parametric error model instead and learn its parameters."""

        if xtrain.shape[0] != ytrain.shape[0]:
            raise TypeError("Xtrain and Ytrain do not have same shape.")

        dist_list = []
        for x0 in xtrain:

            # Change basis
            x0 = np.dot(self.L_mat, x0)

            # Get nearest neighbors in original training set
            dist, loc = self.transf_xtree.query(x0, k=k)
            # Weighted density in ball for NN
            # dist = np.array([np.max([1e-15,d]) for d in dist])
            # weight = 1.0/dist
            # dist_list.append(np.sum(weight))
            dist_list.append(np.mean(dist))

        dist_list = np.array(dist_list)
        scalar_error = np.mean(
            np.abs(ytrain), axis=0
        )  # TODO: is this a good error measure?

        def error_model(dist, a, b, c):
            return a * (dist) + b * (dist) ** c

        bestfit, cov = opt.curve_fit(
            error_model,
            dist_list,
            scalar_error,
            # bounds=((0.0,0.0,0.0),(np.inf,np.inf,np.inf)))
            bounds=((0.0, 0.0, 0.0), (1e1, 1e1, 1e1)),
        )

        # print("this is bestfit:", bestfit)

        def new_error_model(xval):
            xval = np.dot(self.L_mat, xval)
            # Get nearest neighbors in original training set
            dist, loc = self.transf_xtree.query(xval, k=k)
            # Mean distance to NN
            dist = np.mean(dist)

            # dist = dist/bestfit[2]

            err_guess = bestfit[0] * dist + bestfit[1] * dist ** bestfit[2]
            # rand_sign = np.random.rand() - 0.5
            # err_guess *= 1.0 if rand_sign>0.0 else -1.0

            return err_guess

        # DEBUG
        # import matplotlib.pyplot as plt  # type: ignore
        # plt.plot(dist_list, np.abs(ytrain),'bo')
        # plt.plot(dist_list, map(new_error_model,xtrain),'ro')
        # plt.show()

        return new_error_model


def polynomial_fit(
    x_0: np.ndarray, x_nn: np.ndarray, y_nn: np.ndarray, weights: np.ndarray, delta: int
) -> np.ndarray:
    """
    Compute polynomial fit

    Assumes that the point of interest is at the origin

    Parameters
    ----------

    x_0 : np.ndarray
        point at which to evaluate polynomial fit

    x_nn : np.ndarray
        distances to nearest neighbours

    y_nn : np.ndarray
        y-values at nearest neighbours

    delta : int
        degree of polynomial
    """

    n = x_0.shape[0]
    k = x_nn.shape[0]
    m = y_nn.shape[1]

    # Check that shapes are consistent
    assert x_nn.shape == (k, n)
    assert y_nn.shape == (k, m)
    assert weights.shape == (k,)

    N_coefs = n_coef(n, delta)

    assert k >= 2 * N_coefs

    W = np.diagflat(weights)

    Z = np.empty((k, N_coefs), dtype=float)
    for i in range(k):
        Z[i] = all_monomials(x_nn[i, :] - x_0, delta)

    mat = np.linalg.inv(Z.T @ W @ Z) @ Z.T @ W
    assert mat.shape == (N_coefs, k)

    out = np.empty(m)

    for j in range(m):
        pj = mat @ y_nn[:, j]
        out[j] = pj[0]

    return out


def n_coef(n: int, delta: int) -> int:
    """
    Number of monomials in n variables, up to degree delta

    n : int
        Number of variables

    delta : int
        Maximum degree
    """

    out = 0
    for i in range(0, delta + 1):
        out += special.comb(n + i - 1, i)

    return int(out)


def all_monomials(x: np.ndarray, delta: int) -> np.ndarray:
    """
    Construct vector of monomials constructed from x

    Parameters
    ----------

    x : np.ndarray
        Point at which to evaluate monomial

    delta : int
        Maximum degree of monomial
    """

    n = x.shape[0]

    N_coefs = n_coef(n, delta)

    monomial_list = []

    monomial_list.append(np.ones(1))
    for deg in range(1, delta + 1):
        monomial_list.append(monomials_deg(x, deg))

    out = np.concatenate(monomial_list)
    assert out.shape == (N_coefs,)

    return out


def monomials_deg(x: np.ndarray, deg: int) -> np.ndarray:
    """
    Construct all monomials of degree equal to deg

    This brute force implementation is minimally efficient.

    Parameters
    ----------

    x : np.ndarray
        Vector from which monomials should be constructed

    deg : int
        Degree of output monomials
    """

    n = x.shape[0]

    n_terms = int(special.comb(n + deg - 1, deg))

    exponents_list = []

    # Iterate over all sets of exponents in a larger set than needed and pick
    # out the right ones
    # This is deterministic
    for e in product(range(0, deg + 1), repeat=n):
        if np.array(e).sum() == deg:
            exponents_list.append(np.array(e))

    exponents = np.array(exponents_list)
    assert exponents.shape == (n_terms, n)

    out = np.power(x, exponents).prod(axis=1)

    return out
